<?php

/**
 * API response class
 *
 * This class is responsible for formatting and sending a
 * response back to the client.
 *
 * Extends Exception, to allow us to take a short-cut out of deep call
 * hierarchies. This is really for Api_Error_Response's benefit, since
 * PHP neither supports multiple inheritance, or has an exception interface.
 */
class Api_Response extends Exception
{
	protected $type = false; /**< The object type we refer to */
	protected $title = false; /**< Something pretty and describing, usually object name or type */
	/**
	 * Constructor
	 * @param $body Response body
	 * @param $code HTTP response code
	 */
	function __construct($body, $code = 200)
	{
		$this->body = $body;
		$this->code = $code;
	}

	/**
	 * Serialize output
	 *
	 * @param $output_format string
	 * @return string
	 */
	protected function _get_adjusted_response($output_format)
	{
		switch($output_format) {
			case 'xml':
				$content_type = 'application/xml';
				$body = $this->xml_encode($this->body)->saveXML();
				break;
			case 'html':
				$content_type = 'text/html';
				$body = $this->html_encode($this->body);
				if(0 !== strpos($body, '<!DOCTYPE')) {
					$body = sprintf("<!DOCTYPE html>
<html><head>
<link type='text/css' rel='stylesheet' href='/api/css/reset.css' />
<link type='text/css' rel='stylesheet' href='/api/css/monitor.css' />
<title>%s</title>
</head><body>
<div id='container'>
	<div id='sidebar'>
		<div id='monitor'>op5 Monitor</div>
		<p>HTTP API</p>
		<ul>
			<li><a href='/api/help/config'>Config</a></li>
			<li><a href='/api/help/status'>Status</a></li>
		</ul>
		<p>This is the op5 Monitor status API. Each Nagios object can be queried by using the following documentation.</p>
	</div>
	<div id='main'>
		<h1>%s</h1>
		%s
		<h2>Getting help</h2>
		<p>Please visit <a href='/api/help/config'>/api/help/config</a> or <a href='/api/help/status'>/api/help/status</a>.</p>
	</div>
</div>
</body></html>",
					"op5 Monitor API - $this->title",
					$this->title,
					$body);
				}
				break;
			case 'json':
			default:
				$content_type = 'application/json';
				$body = json_encode($this->body);
				break;
		}
		return $body;
	}

	/**
	 * Send a response
	 *
	 * This function takes a request object to find output format from, but
	 * in case of errors parsing the input, it needs to be optional.
	 *
	 * @param $request Api_Request = null
	 * @return string
	 */
	function send(Api_Request $request=null)
	{
		$output_format = false;
		if ($request) {
			$output_format = $request->get_output_format();
			$path = $request->get_path();
			$this->type = reset($path);
			$this->title = end($path);
		}

		$status_response = $this->code;
		switch($this->code) {
			case 200:
				$status_response .= " OK";
				break;
			case 201:
				$status_response .= " Created";
				break;
			case 400:
				$status_response .= " Bad request";
				break;
			case 401:
				$status_response .= " Unauthorized";
				break;
			case 403:
				$status_response .= " Forbidden";
				break;
			case 404:
				$status_response .= " Not Found";
				break;
			case 405:
				$status_response .= " Method Not Allowed";
				break;
			case 500:
				$status_response .= " Internal Server Error";
				break;
			case 501:
				$status_response .= " Not Implemented";
				break;
		}
		header("HTTP/1.0 $status_response");
		switch($output_format) {
		case 'xml':
			$content_type = 'application/xml';
			break;
		case 'html':
			$content_type = 'text/html';
			break;
		case 'json':
		default:
			$content_type = 'application/json';
			break;
		}
		header("Content-Type: $content_type; charset=utf-8");
		return $this->_get_adjusted_response($output_format);
	}

	private function xml_encode_worker($k, $v, $parent, $doc, $elem_name="value") {
		if (!is_numeric($k))
			$parent->setAttribute('name', $k);
		if (is_object($v)) {
			$v = get_object_vars($v);
		}
		if (is_array($v)) {
			foreach ($v as $vk => $vv) {
				$value = $doc->createElement($elem_name);
				$parent->appendChild($value);
				$this->xml_encode_worker($vk, $vv, $value, $doc, "value");
			}
		}
		else {
			$parent->appendChild($doc->createTextNode($v));
		}
	}

	/**
	 * We need to do the seemingly annoying<key>$key</key><value>$value</value>
	 * encoding, because timeperiod exceptions and variables such as 2dcoords
	 * are not valid as XML keys
	 */
	private function xml_encode($data, $root_name='root') {
		$doc = new DOMDocument();
		$doc->formatOutput = true;
		$root = $doc->createElement($root_name);
		$doc->appendChild($root);
		$this->xml_encode_worker(0, $data, $root, $doc, "param");
		return $doc;
	}

	/**
	 * @param $data mixed
	 * @param $escape boolean = false, not used for the whole body but used when listing objects' properties
	 */
	private function html_encode($data, $escape = false) {
		if(is_string($data)) {
			if (empty($data))
				return '[empty]';
			if($escape) {
				return htmlspecialchars($data);
			}
			return $data;
		}
		if(is_numeric($data)) {
			return (string) $data;
		}
		$string = null;
		if(is_array($data)) {
			if (!empty($data) && (array_keys($data) !== range(0, count($data) - 1))) {
				// Associative map (copied from Zend's json encoding)
				$string .= '<dl>';
				foreach($data as $key => $value) {
					$string .= "<dt>".htmlspecialchars($key)."</dt><dd>".$this->html_encode($value, true)."</dd>";
				}
				$string .= '</dl>';
			} else {
				// Array
				$string .= '<ul>';
				foreach($data as $value) {
					$string .= "<li>".$this->html_encode($value, true)."</li>";
				}
				$string .= '</ul>';
			}
		}
		if($data instanceof stdClass) {
			return $this->html_encode((array) $data);
		}
		return $string;
	}
}
